import json
import os
import sys

import click

import storyscript

import xerox

from .utils import find_story_yml, get_app_name_from_yml, get_asyncy_yaml
from .version import compiler_version, version


def echo_support():

    data = {}
    data['cli'] = {}
    data['compiler'] = {}
    data['python'] = {}
    data['app'] = {}
    data['context'] = {}

    data['cli']['version'] = version
    data['cli']['path'] = os.path.dirname(__file__)
    data['compiler']['version'] = compiler_version
    data['compiler']['path'] = os.path.dirname(storyscript.__file__)
    data['python']['path'] = sys.executable
    data['python']['version'] = sys.version.split()[0]
    data['app']['name'] = get_app_name_from_yml()
    data['app']['story.yml'] = get_asyncy_yaml(must_exist=False)
    data['app']['story.yml_path'] = find_story_yml()
    data['context']['cwd'] = os.getcwd()
    data['context']['environ_keys'] = [k for k in os.environ.keys()]
    data['context']['argv'] = sys.argv

    report = json.dumps(data, indent=4)
    click.echo(report)

    # Copy the report to the clipboard!
    try:
        xerox.copy(report)
        click.echo(click.style('Copied to clipboard!', fg='red', bold=True))
    except Exception:
        pass

    # TODO: this.



# -*- coding: utf-8 -*-
import time

from pytest import mark


@mark.parametrize('with_message', [True, False])
@mark.parametrize('hard_deployment', [True, False])
@mark.parametrize('final_release_state', [
    'DEPLOYED', 'FAILED', 'UNKNOWN', 'TEMP_DEPLOYMENT_FAILURE'
])
@mark.parametrize('maintenance', [True, False])
@mark.parametrize('payload', [
    None, {'stories': {'foo'}, 'services': ['bar', 'baz']}
])
def test_deploy(runner, with_message, patch, hard_deployment,
                final_release_state, maintenance, payload,
                init_sample_app_in_cwd):
    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()

        # Relative imports are used here since we need to trigger
        # the cli init code in an isolated filesystem, inside an app dir.
        # Weird things happen otherwise. Not the most efficient way, but works.
        from story import api
        from story.commands import test
        from story.commands.deploy import deploy

        patch.object(test, 'compile_app', return_value=payload)
        patch.object(time, 'sleep')

        patch.object(api.Config, 'get')
        patch.object(api.Releases, 'create')
        patch.object(api.Releases, 'get', side_effect=[
            [{'state': 'QUEUED'}],
            [{'state': 'DEPLOYING'}],
            [{'state': final_release_state}],
        ])
        patch.object(api.Apps, 'maintenance', return_value=maintenance)

        args = []

        if with_message:
            message = 'hello world'
            args.append('--message')
            args.append(message)
        else:
            message = None

        if hard_deployment:
            args.append('--hard')

        if payload is None:
            result = runner.run(deploy, exit_code=1)
            assert result.stdout == ''
            return
        else:
            result = runner.run(deploy, exit_code=0, args=args)

        if maintenance:
            assert 'Your app is in maintenance mode.' in result.stdout
            return

        api.Config.get.assert_called_with('my_app')
        api.Releases.create.assert_called_with(
            api.Config.get(), payload, 'my_app', message, hard_deployment)

        assert time.sleep.call_count == 3

        if final_release_state == 'DEPLOYED':
            assert 'Configured 1 story' in result.stdout
            assert '- foo' in result.stdout
            assert 'Deployed 2 services' in result.stdout
            assert '- bar' in result.stdout
            assert '- baz' in result.stdout
            assert 'Created ingress route' in result.stdout
            assert 'Configured logging' in result.stdout
            assert 'Configured health checks' in result.stdout
            assert 'Deployment successful!' in result.stdout
        elif final_release_state == 'FAILED':
            assert 'Deployment failed!' in result.stdout
            assert 'story logs' in result.stdout
        elif final_release_state == 'TEMP_DEPLOYMENT_FAILURE':
            assert 'Deployment failed!' in result.stdout
            assert 'status.storyscript.io' in result.stdout
        else:
            assert f'An unhandled state of your app has been encountered ' \
                   f'- {final_release_state}' in result.stdout
            assert 'support@storyscript.io' in result.stdout



import click

import requests

import semver

from . import storage
from .version import version as story_version

PYPI_API_URL = 'https://pypi.org/pypi/story/json'
CACHE_EXPIRES = 60 * 60 * 3
REQUEST_TIMEOUT = 2

http_session = requests.Session()


def _latest_pypi():
    # Restore from cache, if applicable.
    if 'cli-latest' in storage.cache:
        return storage.cache['cli-latest']

    # Make an HTTP request to the PyPI JSON API.
    try:
        r = http_session.get(url=PYPI_API_URL, timeout=REQUEST_TIMEOUT)
        r.raise_for_status()
    except Exception:
        return None

    # Fetch the releases from the JSON response.
    releases = r.json()['releases']

    # Grab the latest release.
    latest = [k for k in releases.keys()][-1]

    # Store the results in the cache for three hours.
    storage.cache.store('cli-latest', latest, expires=CACHE_EXPIRES)

    # Parse the version string.
    return latest


def ensure_latest():
    current_version = '.'.join(story_version.split('.')[:3])

    # Grab the latest version from PyPi. Valid for three hours.
    latest_version = _latest_pypi()

    if latest_version:
        if semver.cmp(current_version, latest_version) == -1:
            click.echo(
                click.style(
                    f'A new release (v{latest_version}) of the '
                    f'Storyscript CLI is now available!',
                    fg='yellow',
                ),
                err=True,
            )



# -*- coding: utf-8 -*-

import sys
from json import dumps
from urllib.error import URLError

import click

import requests

from . import cli, storage
from .environment import SS_GRAPHQL


def graphql(query, **variables):
    try:
        res = requests.post(
            SS_GRAPHQL,
            data=dumps({'query': query, 'variables': variables}),
            headers={
                'Content-Type': 'application/json',
                'Accept': 'application/json',
                'Authorization': f'Bearer {cli.get_access_token()}',
            },
            timeout=10,
        )
    except KeyboardInterrupt:  # OK - user cancelled.
        click.echo('\nCancelled')
        click.get_current_context().exit(1)
        return
    except (URLError, requests.RequestException):
        click.echo(
            click.style(f'\nFailed to connect to {SS_GRAPHQL}', fg='red'),
            err=True,
        )
        click.get_current_context().exit(1)
        return

    data = res.json()
    if 'errors' in data:
        click.echo()
        for error in data['errors']:
            if error['message'] == 'InvalidOrExpiredToken':
                click.echo('Your session has expired.', err=True)
                click.echo()
                cli.reset()
            else:
                click.echo(
                    click.style('Error: ', fg='red') + error['message'],
                    err=True
                )
        click.get_current_context().exit(1)
        return
    return data


class Config:
    @staticmethod
    def get(app: str):
        res = graphql(
            """
            query($app: UUID!){
              allReleases(condition: {appUuid: $app},
                          first: 1, orderBy: ID_DESC){
                nodes{
                  config
                }
              }
            }
            """,
            app=Apps.get_uuid_from_hostname(app),
        )
        try:
            return res['data']['allReleases']['nodes'][0]['config'] or {}
        except:
            return {}

    @staticmethod
    def set(config: {}, app: str, message: str) -> dict:
        return Releases.create(
            config, None, app, message or 'Update environment', False
        )


class Releases:
    @staticmethod
    def list(app: str, limit: int):
        res = graphql(
            """
            query($app: UUID!){
              allReleases(condition: {appUuid: $app},
                        first: 30, orderBy: ID_DESC){
                nodes{
                  id
                  message
                  timestamp
                  state
                }
              }
            }
            """,
            app=Apps.get_uuid_from_hostname(app),
        )
        try:
            return res['data']['allReleases']['nodes']
        except:
            return []

    @staticmethod
    def rollback(version: str, app: str):
        app_uuid = Apps.get_uuid_from_hostname(app)
        res = graphql(
            """
            query($app: UUID!, $version: Int!){
              releaseByAppUuidAndId( appUuid: $app, id: $version){
                config
                payload
              }
            }
            """,
            app=app_uuid,
            version=version,
        )
        release = res['data']['releaseByAppUuidAndId']

        return Releases.create(
            release['config'] or {},
            release['payload'],
            app,
            f'Rollback to v{version}',
            False
        )

    @staticmethod
    def get(app: str):
        res = graphql(
            """
            query($app: UUID!){
              allReleases(condition: {appUuid: $app},
                          first: 1, orderBy: ID_DESC){
                nodes{
                  id
                  state
                }
              }
            }
            """,
            app=Apps.get_uuid_from_hostname(app),
        )
        try:
            return res['data']['allReleases']['nodes']
        except:
            return []

    @staticmethod
    def create(config: {}, payload: {}, app: str, message: str,
               always_pull_images: bool) -> dict:
        res = graphql(
            """
            mutation ($data: CreateReleaseInput!){
              createRelease(input: $data) {
                release { id }
              }
            }
            """,
            data={
                'release': {
                    'appUuid': Apps.get_uuid_from_hostname(app),
                    'message': message or 'Deploy app',
                    'config': config,
                    'payload': payload,
                    'alwaysPullImages': always_pull_images
                }
            },
        )

        release_ = res['data']['createRelease']['release']

        changes = 'Code'
        if payload is None:
            changes = 'Config'

        cli.track(
            'App Release Created',
            {'Version': release_['id'], 'App name': app, 'Changes': changes},
        )

        return release_


class Apps:
    _hostname_to_uuid = {}

    @staticmethod
    def get_uuid_from_hostname(app: str) -> str:
        """
        This method also caches all queries made, but only during the current
        CLI session.
        """
        cache_key = f'app-{app}'
        the_uuid = None

        if cache_key in storage.cache:
            the_uuid = storage.cache.fetch(cache_key)

        if not the_uuid:
            the_uuid = Apps._hostname_to_uuid.get(app)

        # Cache the app's UUID.
        if cache_key not in storage.cache:
            storage.cache.store(cache_key, the_uuid)

        res = graphql(
            """
            query($app: Hostname!){
              app: appDnsByHostname(hostname: $app){
                appUuid
              }
            }
            """,
            app=app,
        )
        if res['data']['app'] is None:
            click.echo()
            click.echo(
                click.style(
                    f'The app "{app}" doesn\'t seem to exist.\n'
                    f'Are you sure you have access to it?',
                    fg='red',
                ),
                err=True,
            )
            sys.exit(1)

        app_uuid = res['data']['app']['appUuid']
        Apps._hostname_to_uuid[app] = app_uuid

        return app_uuid

    @staticmethod
    def maintenance(app: str, maintenance: bool):
        if maintenance is None:
            res = graphql(
                """
                query($app: UUID!){
                  app: appByUuid(uuid: $app){
                    maintenance
                  }
                }
                """,
                app=Apps.get_uuid_from_hostname(app),
            )
            return res['data']['app']['maintenance']
        else:
            graphql(
                """
                mutation ($data: UpdateAppByUuidInput!){
                  updateAppByUuid(input: $data){
                    app{
                      uuid
                    }
                  }
                }
                """,
                data={
                    'uuid': Apps.get_uuid_from_hostname(app),
                    'appPatch': {'maintenance': maintenance},
                },
            )

    @staticmethod
    def list() -> list:
        res = graphql(
            """
            query{
              allApps(condition: {deleted: false}, orderBy: NAME_ASC){
                nodes{
                  name
                  timestamp
                  maintenance
                }
              }
            }
            """
        )
        return res['data']['allApps']['nodes']

    @staticmethod
    def create(name: str, team: str) -> dict:
        res = graphql(
            """
            mutation ($data: CreateAppInput!){
              createApp(input: $data) {
                app{
                  name
                }
              }
            }
            """,
            data={'app': {'ownerUuid': cli.get_user_id(), 'name': name}},
        )
        return res['data']['createApp']['app']

    @staticmethod
    def destroy(app: str):
        graphql(
            """
            mutation ($data: UpdateAppByUuidInput!){
              updateAppByUuid(input: $data){
                app{
                  uuid
                }
              }
            }
            """,
            data={
                'uuid': Apps.get_uuid_from_hostname(app),
                'appPatch': {'deleted': True},
            },
        )



__all__ = [
    'apps',
    'template',
    'config',
    'completion',
    'feedback',
    'login',
    'logs',
    'maintenance',
    'releases',
    'status',
    'test',
    'update',
    'deploy',
    'logout'
]



# -*- coding: utf-8 -*-

import click

from .. import cli


@cli.cli.command()
def feedback():
    """Give feedback."""

    click.echo('Opening https://asyncy.click/feedback')
    click.launch('https://asyncy.click/feedback')



# -*- coding: utf-8 -*-
import json
import os

import click

from pytest import mark

from story import cli

from storyscript.App import App
from storyscript.exceptions import StoryError


@mark.parametrize('mock_tree_as_none', [True, False])
@mark.parametrize('no_stories_found', [True, False])
@mark.parametrize('debug', [True, False])
def test_test(runner, patch, init_sample_app_in_cwd, mock_tree_as_none,
              no_stories_found, debug):
    expected_exit_code = 0
    compile_app_result = {
        'stories': {
            'a1.story': {},
            'a2.story': {}
        }
    }

    args = []
    if debug:
        args.append('--debug')

    if no_stories_found:
        compile_app_result['stories'] = {}
        expected_exit_code = 1

    if mock_tree_as_none:
        compile_app_result = None
        expected_exit_code = 1

    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()
        from story.commands import test
        patch.object(test, 'compile_app', return_value=compile_app_result)

        result = runner.run(test.test, args=args, exit_code=expected_exit_code)

    test.compile_app.assert_called_with('my_app', debug)

    if mock_tree_as_none:
        assert result.stdout == ''
    elif no_stories_found:
        assert 'No stories found' in result.stdout
    else:
        assert 'Looking good!' in result.stdout
        assert 'a1.story' in result.stdout
        assert 'a2.story' in result.stdout


@mark.parametrize('force_compilation_error,compilation_exc', [
    (True, StoryError('E100', 'a.story', path='foo')),
    (True, BaseException('oh no!')),
    (False, None)
])
@mark.parametrize('nested_dir', [True, False])
def test_compile_app(runner, patch, init_sample_app_in_cwd,
                     pre_init_cli_runner, nested_dir,
                     force_compilation_error, compilation_exc):
    app_name_for_analytics = 'my_special_app'

    patch.object(cli, 'get_asyncy_yaml', return_value='asyncy_yml_content')
    patch.object(cli, 'track')

    patch.object(click, 'echo')

    if force_compilation_error:
        patch.object(App, 'compile',
                     side_effect=compilation_exc)

    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()
        actual_project_root = os.getcwd()
        pre_init_cli_runner()
        from story.commands import test

        # nested_dir is used to specifically ensure that when compile_app
        # is executed in a sub directory under the main project, all stories
        # in the project are compiled, and not just the ones that were
        # in that sub directory.
        if nested_dir:
            os.chdir(f'{os.getcwd()}/src/b')

        actual_compilation_result = test.compile_app(app_name_for_analytics,
                                                     False)

        if not force_compilation_error:
            os.chdir(actual_project_root)
            expected_compilation_result = json.loads(App.compile(os.getcwd()))
            expected_compilation_result['yaml'] = 'asyncy_yml_content'

    # Ugly assert, I know. Can't help it since we're not running this via
    # the runner.
    assert 'Compiling Stories' in click.echo.mock_calls[0][1][0]

    if force_compilation_error:
        assert actual_compilation_result is None
        assert 'Failed to compile project' in click.echo.mock_calls[1][1][0]
        err_line = click.echo.mock_calls[2][1][0]
        if isinstance(compilation_exc, StoryError):
            assert 'Please report at ' \
                   'https://github.com/storyscript' \
                   '/storyscript/issues' in err_line
        else:
            assert str(compilation_exc) in err_line
    else:
        assert actual_compilation_result == expected_compilation_result

    cli.track.assert_called_with('App Compiled', {
        'App name': app_name_for_analytics,
        'Result': 'Failed' if force_compilation_error else 'Success',
        'Stories': 0 if force_compilation_error else 2
    })



# -*- coding: utf-8 -*-
import sys

try:
    # Allow keyboard interrupts of CLI import.
    from .cli import cli

    # Import commands for CLI.
    from .commands import *  # noqa

except KeyboardInterrupt:
    print('Aborted!')
    sys.exit(1)

if __name__ == '__main__':
    try:
        cli()
    except KeyboardInterrupt:
        print('Aborted!')
        sys.exit(1)



# -*- coding: utf-8 -*-

import click


try:
    from . import cli

    _app = cli.get_app_name_from_yml()
except Exception:
    _app = None


def app(allow_option=True):
    return click.option(
        '--app',
        '-a',
        default=_app,
        help=f'App to interact with [default: {_app}].',
        callback=lambda context, p, app: cli.assert_project(
            context.command.name, app, _app, allow_option
        ),
    )



# -*- coding: utf-8 -*-
import click

from story.commands.feedback import feedback


def test_feedback(runner, patch):
    patch.object(click, 'launch')

    with runner.runner.isolated_filesystem():
        result = runner.run(feedback, exit_code=0)

    feedback_url = 'https://asyncy.click/feedback'
    click.launch.assert_called_with(feedback_url)
    assert feedback_url in result.stdout



# -*- coding: utf-8 -*-
from unittest import mock

from pytest import mark

from story.helpers import datetime


@mark.parametrize('no_releases', [True, False])
@mark.parametrize('limit', [None, 100, 200])
def test_list(runner, patch, init_sample_app_in_cwd, no_releases, limit):
    if not no_releases:
        patch.object(datetime, 'reltime', side_effect=[
            'reltime_100',
            'reltime_200',
            'reltime_300'
        ])

        patch.object(datetime, 'parse_psql_date_str', side_effect=[
            'parsed_100',
            'parsed_200',
            'parsed_300'
        ])

    args = []
    if limit is not None:
        args.append('-n')
        args.append(limit)

    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()

        from story import api
        from story.commands.releases import list_command

        if no_releases:
            patch.object(api.Releases, 'list', return_value=[])
        else:
            patch.object(api.Releases, 'list', return_value=[
                {
                    'id': 100,
                    'state': 'deployed',
                    'timestamp': 'date_app_100',
                    'message': 'my_deployment_message_100'
                },
                {
                    'id': 300,
                    'state': 'deployed',
                    'timestamp': 'date_app_300',
                    'message': 'my_deployment_message_300'
                },
                {
                    'id': 200,
                    'state': 'deployed',
                    'timestamp': 'date_app_200',
                    'message': 'my_deployment_message_200'
                },
            ])

        result = runner.run(list_command, args=args, exit_code=0)

    if limit is None:
        limit = 20  # The supposed default.

    api.Releases.list.assert_called_with('my_app', limit=limit)

    if no_releases:
        assert 'No releases yet for app my_app' in result.stdout
    else:
        assert """
    VERSION    STATUS      CREATED              MESSAGE         
============================================================
v100      Deployed   reltime_100   my_deployment_message_100
v200      Deployed   reltime_200   my_deployment_message_200
v300      Deployed   reltime_300   my_deployment_message_300
""".strip() in result.stdout  # noqa (because there's a trailing whitespace in the header)

        assert datetime.parse_psql_date_str.mock_calls == [
            mock.call('date_app_100'),
            mock.call('date_app_200'),
            mock.call('date_app_300')
        ]

        assert datetime.reltime.mock_calls == [
            mock.call('parsed_100'),
            mock.call('parsed_200'),
            mock.call('parsed_300')
        ]


@mark.parametrize('version_specified,version_number', [
    (False, None),
    (True, '28'),
    (True, '0'),
    (True, 'v28'),
    (True, 'invalid')
])
def test_rollback(runner, patch, init_sample_app_in_cwd,
                  version_specified, version_number):
    expected_rollback_version = 99
    expected_exit_code = 0

    args = []

    if version_specified:
        expected_rollback_version = version_number

        if expected_rollback_version[0] == 'v':
            expected_rollback_version = expected_rollback_version[1:]

        args.append(version_number)

        if version_number == '0' or version_number == 'invalid':
            expected_exit_code = 1

    with runner.runner.isolated_filesystem():
        init_sample_app_in_cwd()

        from story import api
        from story.commands.releases import rollback

        if not version_specified:
            patch.object(api.Releases, 'get', return_value=[{'id': 100}])
        else:
            patch.object(api.Releases, 'get',
                         return_value=[{'id': version_number}])

        patch.object(api.Releases, 'rollback',
                     return_value={'id': expected_rollback_version})

        result = runner.run(rollback, exit_code=expected_exit_code, args=args)

    if version_number == 'invalid':
        assert 'Invalid release specified.' in result.stdout
        return

    if version_number is not None and version_number[0] == 'v':
        version_number = version_number[1:]

    if version_specified and int(version_number) <= 0:
        assert 'Unable to rollback a release before v1' in result.stdout
        return

    if not version_specified:
        api.Releases.get.assert_called_with(app='my_app')
        assert 'Getting latest release for app my_app' in result.stdout

    assert f'Rolling back to v{expected_rollback_version}' in result.stdout
    assert 'Deployed new release' in result.stdout

    api.Releases.rollback.assert_called_with(version=expected_rollback_version,
                                             app='my_app')



# -*- coding: utf-8 -*-

"""
    Copied from https://github.com/click-contrib/click-didyoumean

    Extension for the python ``click`` module to provide
    a group with a git-like *did-you-mean* feature.
"""

import difflib

import click


__version__ = '0.0.3'


class DYMMixin(object):  # pylint: disable=too-few-public-methods
    """
    Mixin class for click MultiCommand inherited classes
    to provide git-like *did-you-mean* functionality when
    a certain command is not registered.
    """

    def __init__(self, *args, **kwargs):
        self.max_suggestions = kwargs.pop('max_suggestions', 3)
        self.cutoff = kwargs.pop('cutoff', 0.5)
        super(DYMMixin, self).__init__(*args, **kwargs)

    def resolve_command(self, ctx, args):
        """
        Overrides clicks ``resolve_command`` method
        and appends *Did you mean ...* suggestions
        to the raised exception message.
        """
        try:
            return super(DYMMixin, self).resolve_command(ctx, args)
        except click.exceptions.UsageError as error:
            error_msg = str(error)
            original_cmd_name = click.utils.make_str(args[0])
            matches = difflib.get_close_matches(
                original_cmd_name,
                map(lambda c: c.replace('-', ':'), self.list_commands(ctx)),
                self.max_suggestions,
                self.cutoff,
            )
            if matches:
                matches = '\n    '.join(
                    map(lambda m: click.style(m, fg='magenta'), matches)
                )
                error_msg += f'\n\nDid you mean one of these?\n    {matches}'

            raise click.exceptions.UsageError(error_msg, error.ctx)


class DYMGroup(DYMMixin, click.Group):
    """
    click Group to provide git-like
    *did-you-mean* functionality when a certain
    command is not found in the group.
    """


class DYMCommandCollection(DYMMixin, click.CommandCollection):
    """
    click CommandCollection to provide git-like
    *did-you-mean* functionality when a certain
    command is not found in the group.
    """
